#ifndef WRAPPER_H
#define WRAPPER_H

#include <GL/glew.h>
#include <GLFW/glfw3.h>
#include "shader/ShaderProgram.h"
#include "../math/Vector3D.h"
#include "../math/Matrix3D.h"
#include "shader/WorldShader.h"
#include "shader/SSAOShader.h"
#include "shader/SSAOBlurShader.h"
#include "shader/FramebufferRectangle.h"
#include "shader/WorldPostShader.h"
#include "shader/OverlayShader.h"
#include "Mesh.h"
#include <iostream>

using namespace std;

class IClient
{
public:
    virtual void tick() = 0;
    virtual void render3DTick(float lag) = 0;
    virtual void render2DTick(float lag) = 0;
    virtual void onKeyEvent(int key, int scancode, int action, int mods) = 0;
    virtual void onMouseClick(int button, int action, int mods) = 0;
};

class DummyClient : public IClient
{
public:
    static DummyClient dummy;
    
    void tick() override { cout << "Dummy tick" << endl; };
    void render3DTick(float lag) override { cout << "Dummy render3DTick" << endl; };
    void render2DTick(float lag) override { cout << "Dummy render3DTick" << endl; };
    void onKeyEvent(int key, int scancode, int action, int mods) override { cout << "Dummy onKeyEvent" << endl; };
    void onMouseClick(int button, int action, int mods) override { cout << "Dummy onMouseClick" << endl; };
private:
    DummyClient() {};
};

class Engine
{
public:
    static bool init(int width, int height, const char* name);
    static void start(IClient* client);
    static void stop();
    
    static void printError();
    
    static int getScale();
    static int getWidth();
    static int getHeight();
    
    static float getFieldOfView();
    static float getNearClip();
    static float getFarClip();
    
    static void setWorldViewMatrix(const float* data);
    static void setWorldModelMatrix(const float* data);
    static void setOverlayModelMatrix(const float* data);

private:
    static const uint64_t NANOS_PER_TICK = 50000000;
    static const int MAX_TICKS_PER_FRAME = 5;
    
    static bool initOverlayShader();   
    
    static void onRenderTick(float lag);
    
    static void onKeyEvent(GLFWwindow* w, int key, int scancode, int action, int mods);
    static void onMouseClick(GLFWwindow* w, int button, int action, int mods);
    static void onWindowResize(GLFWwindow* w, int width, int height);
    
    static void updateScale();
    
    static IClient* client;    
    
    // window data
    static GLFWwindow* window;  
    static int scale;
    static int width;
    static int height;
    
    // projection data
    static float fovY;
    static float nearClip;
    static float farClip;
    static Matrix3D projMatrix;
    
    // rectangle for framebuffer drawing
    static FramebufferRectangle rectangle;
    
    // shader stage 1 - world
    static WorldShader worldShader;
    // shader stage 2 - world ssao
    static SSAOShader ssaoShader;
    // shader stage 3 - world ssao blur
    static SSAOBlurShader ssaoBlurShader;
    // shader stage 4 - world post
    static WorldPostShader worldPostShader;
    // shader stage 5 - 2D overlay
    static OverlayShader overlayShader;
};

#endif